from functools import partial
from typing import Any

import numpy as np
import pytest
from numpy.typing import NDArray

from ropt.plan import BasicOptimizer
from ropt.plugins.realization_filter.default import (
    _get_cvar_weights_from_percentile,
    _sort_and_select,
)
from ropt.results import FunctionResults, GradientResults, Results


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> dict[str, Any]:
    return {
        "optimizer": {
            "speculative": True,
            "max_functions": 10,
        },
        "objectives": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "number_of_perturbations": 5,
            "perturbation_magnitudes": 0.01,
        },
        "realizations": {
            "weights": 5 * [1.0],
        },
        "variables": {
            "initial_values": 3 * [0],
        },
    }


@pytest.mark.parametrize(
    ("failed", "first", "last", "expected"),
    [
        ([False, False, False, False, False], 0, 1, [0, 0, 0, 0.4, 0.5]),
        ([False, False, False, False, False], 1, 2, [0, 0, 0.3, 0.4, 0]),
        ([False, False, False, True, False], 0, 1, [0, 0, 0.3, 0, 0.5]),
        ([False, False, True, False, False], 1, 2, [0.1, 0, 0.0, 0.4, 0]),
    ],
)
def test__sort_and_select(
    failed: list[bool], first: int, last: int, expected: list[float]
) -> None:
    values = np.array([3, 4, 2, 1, 0])
    weights = np.array([0.1, 0.2, 0.3, 0.4, 0.5])
    result = _sort_and_select(values, weights, np.array(failed), first, last)
    assert np.allclose(result, expected)


@pytest.mark.parametrize(
    ("failed", "percentile", "weights"),
    [
        ([False, False, False, False, False], 0.1, [0.0, 0.0, 0.0, 0.0, 0.1]),
        ([False, False, False, False, False], 0.2, [0.0, 0.0, 0.0, 0.0, 0.2]),
        ([False, False, False, False, False], 0.3, [0.0, 0.0, 0.0, 0.1, 0.2]),
        ([False, False, False, False, False], 0.35, [0.0, 0.0, 0.0, 0.15, 0.2]),
        ([False, False, False, False, False], 0.4, [0.0, 0.0, 0.0, 0.2, 0.2]),
        ([False, False, False, False, False], 0.9, [0.2, 0.1, 0.2, 0.2, 0.2]),
        ([False, False, False, False, False], 1.0, [0.2, 0.2, 0.2, 0.2, 0.2]),
        ([False, False, False, True, False], 0.4, [0, 0, 0.15, 0.0, 0.25]),
    ],
)
def test__get_cvar_weights_from_percentile(
    failed: list[bool], percentile: float, weights: list[float]
) -> None:
    values = np.array([3, 4, 2, 1, 0])
    estimated_weights = _get_cvar_weights_from_percentile(
        values,
        np.array(failed),
        percentile,
    )
    assert np.allclose(estimated_weights, weights)


def _objective_function(
    variables: NDArray[np.float64],
    context: Any,
    target: NDArray[np.float64],
) -> float:
    diff: NDArray[np.float64] = variables - target
    # Make sure that some realizations yield a different result:
    if context.realization % 2 == 0:
        diff += 1.0
    result = np.sum(diff**2)
    # Make sure the other realizations have a worse result:
    if context.realization % 2 == 1:
        result += 10.0
    return float(result)


def _constraint_function(variables: NDArray[np.float64], context: Any) -> float:
    # Track how often this function is called.
    result = variables[0] + variables[2]
    # Break some realizations, same as in the distance function:
    if context.realization % 2 == 0:
        result = variables[0] + variables[2] - 10.0
    return float(result)


def _track_results(results: tuple[Results, ...], result_list: list[Results]) -> None:
    result_list.extend(results)


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_sort_filter_on_objectives(
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations

    variables = BasicOptimizer(enopt_config, evaluator(functions)).run().variables
    assert variables is not None
    assert not np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)

    enopt_config["realization_filters"] = [
        {
            "method": "sort-objective",
            "options": {
                "sort": [0],
                "first": 3,
                "last": 4,
            },
        },
    ]
    enopt_config["objectives"]["realization_filters"] = [0, 0]

    result_list: list[Results] = []
    results = (
        BasicOptimizer(
            enopt_config,
            evaluator(functions),
        )
        .set_results_callback(partial(_track_results, result_list=result_list))
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_sort_filter_on_objectives_with_constraints(
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
        partial(_constraint_function),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations

    enopt_config["nonlinear_constraints"] = {
        "lower_bounds": -np.inf,
        "upper_bounds": 0.4,
    }
    enopt_config["realization_filters"] = [
        {
            "method": "sort-objective",
            "options": {
                "sort": [0],
                "first": 3,
                "last": 4,
            },
        },
    ]
    enopt_config["objectives"]["realization_filters"] = [0, 0]
    enopt_config["nonlinear_constraints"]["realization_filters"] = [0]
    result_list: list[Results] = []
    results = (
        BasicOptimizer(
            enopt_config,
            evaluator(functions),
        )
        .set_results_callback(partial(_track_results, result_list=result_list))
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [-0.05, 0.0, 0.45], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            assert result.evaluations.perturbed_constraints is not None
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] != 0.0
                )


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_sort_filter_on_constraints(
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
        partial(_constraint_function),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations
    enopt_config["nonlinear_constraints"] = {
        "lower_bounds": -np.inf,
        "upper_bounds": 0.4,
    }
    enopt_config["realization_filters"] = [
        {
            "method": "sort-constraint",
            "options": {
                "sort": 0,
                "first": 3,
                "last": 4,
            },
        },
    ]
    enopt_config["objectives"]["realization_filters"] = [0, 0]
    enopt_config["nonlinear_constraints"]["realization_filters"] = [0]
    result_list: list[Results] = []
    results = (
        BasicOptimizer(
            enopt_config,
            evaluator(functions),
        )
        .set_results_callback(
            partial(_track_results, result_list=result_list),
        )
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [-0.05, 0.0, 0.45], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            assert result.evaluations.perturbed_constraints is not None
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] != 0.0
                )


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_sort_filter_mixed(  # noqa: C901
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
    ]

    enopt_config["objectives"]["weights"] = [0.75, 0.25, 0.75, 0.25]
    enopt_config["optimizer"]["split_evaluations"] = split_evaluations

    objective_values: list[NDArray[np.float64]] = []

    def _add_objective(results: tuple[Results, ...]) -> None:
        if results:
            for item in results:
                if isinstance(item, FunctionResults):
                    assert item.functions is not None
                    objective_values.append(item.functions.weighted_objective)
        _track_results(results, result_list=result_list)

    # Apply the filtering to all objectives, giving the expected result.
    enopt_config["realization_filters"] = [
        {
            "method": "sort-objective",
            "options": {
                "sort": [0],
                "first": 3,
                "last": 4,
            },
        },
    ]
    enopt_config["objectives"]["realization_filters"] = [0, 0, 0, 0]

    result_list: list[Results] = []
    results = (
        BasicOptimizer(enopt_config, evaluator(functions))
        .set_results_callback(_add_objective)
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )

    # Apply filtering only to the first two, giving a wrong result.
    enopt_config["realization_filters"] = [
        {
            "method": "sort-objective",
            "options": {
                "sort": [0],
                "first": 3,
                "last": 4,
            },
        },
    ]
    enopt_config["objectives"]["realization_filters"] = [0, 0, -1, -1]

    result_list = []
    results = (
        BasicOptimizer(enopt_config, evaluator(functions))
        .set_results_callback(_add_objective)
        .run()
        .results
    )
    assert results is not None
    assert not np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, :, :2] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, :, :2] != 0.0
                )
            assert np.all(
                result.evaluations.perturbed_objectives[filtered, :, 2:] != 0.0
            )

    # The first objective values should differ.
    assert objective_values[0] != objective_values[1]


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_cvar_filter_on_objectives(
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations

    variables = BasicOptimizer(enopt_config, evaluator(functions)).run().variables
    assert variables is not None
    assert not np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)

    enopt_config["realization_filters"] = [
        {
            "method": "cvar-objective",
            "options": {
                "sort": [0],
                "percentile": 0.4,
            },
        },
    ]
    enopt_config["objectives"]["realization_filters"] = [0, 0]
    result_list: list[Results] = []
    results = (
        BasicOptimizer(
            enopt_config,
            evaluator(functions),
        )
        .set_results_callback(
            partial(_track_results, result_list=result_list),
        )
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_cvar_filter_on_objectives_with_constraints(
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
        partial(_constraint_function),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations

    enopt_config["nonlinear_constraints"] = {
        "lower_bounds": -np.inf,
        "upper_bounds": 0.4,
    }
    enopt_config["realization_filters"] = [
        {
            "method": "cvar-objective",
            "options": {
                "sort": [0],
                "percentile": 0.4,
            },
        },
    ]
    enopt_config["objectives"]["realization_filters"] = [0, 0]
    enopt_config["nonlinear_constraints"]["realization_filters"] = [0]
    result_list: list[Results] = []
    results = (
        BasicOptimizer(
            enopt_config,
            evaluator(functions),
        )
        .set_results_callback(
            partial(_track_results, result_list=result_list),
        )
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [-0.05, 0.0, 0.45], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            assert result.evaluations.perturbed_constraints is not None
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] != 0.0
                )


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_cvar_filter_on_constraints(
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
        partial(_constraint_function),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations

    enopt_config["nonlinear_constraints"] = {
        "lower_bounds": -np.inf,
        "upper_bounds": 0.4,
    }
    enopt_config["realization_filters"] = [
        {
            "method": "cvar-constraint",
            "options": {
                "sort": 0,
                "percentile": 0.4,
            },
        },
    ]
    enopt_config["objectives"]["realization_filters"] = [0, 0]
    enopt_config["nonlinear_constraints"]["realization_filters"] = [0]
    result_list: list[Results] = []
    results = (
        BasicOptimizer(
            enopt_config,
            evaluator(functions),
        )
        .set_results_callback(
            partial(_track_results, result_list=result_list),
        )
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [-0.05, 0.0, 0.45], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            assert result.evaluations.perturbed_constraints is not None
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] != 0.0
                )


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_cvar_filter_mixed(  # noqa: C901
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations
    enopt_config["objectives"]["weights"] = [0.75, 0.25, 0.75, 0.25]

    objective_values: list[NDArray[np.float64]] = []

    def _add_objective(results: tuple[Results, ...]) -> None:
        for item in results:
            if isinstance(item, FunctionResults):
                assert item.functions is not None
                objective_values.append(item.functions.weighted_objective)
        _track_results(results, result_list=result_list)

    # Apply the filtering to all objectives, giving the expected result.
    enopt_config["realization_filters"] = [
        {
            "method": "cvar-objective",
            "options": {
                "sort": [0],
                "percentile": 0.4,
            },
        },
    ]
    enopt_config["objectives"]["realization_filters"] = [0, 0, 0, 0]

    result_list: list[Results] = []
    results = (
        BasicOptimizer(enopt_config, evaluator(functions))
        .set_results_callback(_add_objective)
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )

    # Apply filtering only to the first two, giving a wrong result.
    enopt_config["realization_filters"] = [
        {
            "method": "cvar-objective",
            "options": {
                "sort": [0],
                "percentile": 0.4,
            },
        },
    ]
    enopt_config["objectives"]["realization_filters"] = [0, 0, -1, -1]

    result_list = []
    results = (
        BasicOptimizer(enopt_config, evaluator(functions))
        .set_results_callback(_add_objective)
        .run()
        .results
    )
    assert results is not None
    assert not np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, :, :2] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, :, :2] != 0.0
                )
            assert np.all(
                result.evaluations.perturbed_objectives[filtered, :, 2:] != 0.0
            )

    # The first objective values should differ.
    assert objective_values[0] != objective_values[1]



"""Plugin functionality for adding optimization plugins.

Optimization plugins are managed by a
[`PluginManager`][ropt.plugins.PluginManager] object, which returns classes or
factory functions to create objects that implement one or more optimization
methods. These objects must adhere to the
[`Optimizer`][ropt.plugins.optimizer.base.Optimizer] abstract base class.
This abstract base class allows `ropt` to provide the optimizer with the
callback used for evaluating functions and gradients and allows it to be started
from an optimizer step in the optimization plan.

To support the implementation of the optimizer classes, the
[`ropt.plugins.optimizer.utils`][ropt.plugins.optimizer.utils] module provides
some utilities.

By default the [`SciPyOptimizer`][ropt.plugins.optimizer.scipy.SciPyOptimizer]
plugin is installed which provides a number of methods from the
[`scipy.optimize`](https://docs.scipy.org/doc/scipy/tutorial/optimize.html)
package.
"""



"""This module defines the base classes for transforms."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray


class VariableTransform(ABC):
    """Abstract base class for variable transformations.

    This class defines the interface for transforming variables between the
    user-defined domain and the optimizer's internal domain. Concrete
    implementations of this class handle the specific logic for each type of
    transformation.

    When implementing a variable transformation, the following aspects must be
    considered:

    - **Variable Value Transformation:** Mapping variable values between the
      user and optimizer domains. This is achieved by overriding the
      [`to_optimizer`][ropt.transforms.base.VariableTransform.to_optimizer]
      and
      [`from_optimizer`][ropt.transforms.base.VariableTransform.from_optimizer]
      methods.
    - **Perturbation Magnitude Transformation:** Stochastic gradient-based
      algorithms use perturbations with specified magnitudes (see
      [`perturbation_magnitudes`][ropt.config.enopt.GradientConfig]). These
      magnitudes are typically defined in the user domain and must be
      transformed to the optimizer domain using the
      [`magnitudes_to_optimizer`][ropt.transforms.base.VariableTransform.magnitudes_to_optimizer]
      method.
    - **Bound Constraint Difference Transformation:** To report violations of
      variable bounds, the differences between variable values and their
      lower/upper bounds must be transformed from the optimizer domain back
      to the user domain. This is done using the
      [`bound_constraint_diffs_from_optimizer`][ropt.transforms.base.VariableTransform.bound_constraint_diffs_from_optimizer]
      method.
    - **Linear Constraint Transformation:** Linear constraints are generally
      defined by coefficients and right-hand-side values in the user domain.
      These must be transformed to the optimizer domain using the
      [`linear_constraints_to_optimizer`][ropt.transforms.base.VariableTransform.linear_constraints_to_optimizer]
      method.
    - **Linear Constraint Difference Transformation:** To report violations of
      linear constraints, the differences between the linear constraint
      values and their right-hand-side values must be transformed back to the
      user domain. This is done using the
      [`linear_constraints_diffs_from_optimizer`][ropt.transforms.base.VariableTransform.linear_constraints_diffs_from_optimizer]
      method.
    """

    @abstractmethod
    def to_optimizer(self, values: NDArray[np.float64]) -> NDArray[np.float64]:
        """Transform values from the user domain to the optimizer domain.

        This method maps variable values from the user-defined domain to the
        optimizer's internal domain. This transformation might involve scaling,
        shifting, or other operations to improve the optimizer's performance.

        The input `values` may be a multi-dimensional array. It is assumed that
        the last axis of the array represents the variable values. If this is
        not the case, you must adjust the order of the axes before and after
        calling this method.

        Args:
            values: The variable values in the user domain to be transformed.

        Returns:
            The transformed variable values in the optimizer domain.
        """

    @abstractmethod
    def from_optimizer(self, values: NDArray[np.float64]) -> NDArray[np.float64]:
        """Transform values from the optimizer domain to the user domain.

        This method maps variable values from the optimizer's internal domain
        back to the user-defined domain. This transformation reverses any
        scaling, shifting, or other operations that were applied to improve
        the optimizer's performance.

        The input `values` may be a multi-dimensional array. It is assumed that
        the last axis of the array represents the variable values. If this is
        not the case, you must adjust the order of the axes before and after
        calling this method.

        Args:
            values: The variable values in the optimizer domain to be
                transformed.

        Returns:
            The transformed variable values in the user domain.
        """

    @abstractmethod
    def magnitudes_to_optimizer(
        self, values: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Transform perturbation magnitudes to the optimizer domain.

        This method transforms perturbation magnitudes, typically used in
        stochastic gradient-based algorithms, from the user-defined domain to
        the optimizer's internal domain. The transformation ensures that the
        perturbations are applied correctly in the optimizer's space, which may
        have different scaling or units than the user domain.

        For example, if variables are scaled down in the optimizer domain, the
        perturbation magnitudes should also be scaled down proportionally.

        Args:
            values: The perturbation magnitudes in the user domain to be transformed.

        Returns:
            The transformed perturbation magnitudes in the optimizer domain.
        """

    @abstractmethod
    def bound_constraint_diffs_from_optimizer(
        self, lower_diffs: NDArray[np.float64], upper_diffs: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Transform bound constraint differences to the user domain.

        This method transforms the differences between variable values and their
        lower/upper bounds from the optimizer's internal domain back to the
        user-defined domain. These differences are used to report constraint
        violations.

        For example, if variables are scaled in the optimizer domain, the
        differences between the variables and their bounds must be scaled back
        to the user domain to accurately reflect the constraint violations in
        the user's original units.

        Args:
            lower_diffs: The differences between the variable values and their
                lower bounds in the optimizer domain.
            upper_diffs: The differences between the variable values and their
                upper bounds in the optimizer domain.

        Returns:
            A tuple containing the transformed differences.
        """

    def linear_constraints_to_optimizer(
        self,
        coefficients: NDArray[np.float64],
        lower_bounds: NDArray[np.float64],
        upper_bounds: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
        """Transform linear constraints from the user domain to the optimizer domain.

        This method transforms linear constraints, defined by their coefficients
        and right-hand-side bounds, from the user-defined domain to the
        optimizer's internal domain. This is essential to maintain the
        validity of the constraints after variable transformations.

        For instance, if variables are scaled or shifted in the optimizer
        domain, the coefficients and bounds of the linear constraints must be
        adjusted accordingly to ensure the constraints remain consistent.

        The linear constraints are defined by the equation `A * x = b`, where `A`
        is the coefficient matrix, `x` is the variable vector, and `b` represents
        the right-hand-side bounds.

        Args:
            coefficients: The coefficient matrix.
            lower_bounds: The lower bounds on the right-hand-side values.
            upper_bounds: The upper bounds on the right-hand-side values.

        Returns:
            A tuple containing the transformed coefficient matrix and bounds.
        """
        msg = "This transformer does not support linear constraints."
        raise NotImplementedError(msg)

    def linear_constraints_diffs_from_optimizer(
        self, lower_diffs: NDArray[np.float64], upper_diffs: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Transform linear constraint differences to the user domain.

        This method transforms the differences between linear constraint values
        and their lower/upper bounds from the optimizer's internal domain back
        to the user-defined domain. These differences are used to report
        constraint violations.

        For example, if linear constraints are scaled in the optimizer domain,
        the differences between the constraint values and their bounds must be
        scaled back to the user domain to accurately reflect the constraint
        violations in the user's original units.

        Args:
            lower_diffs: The differences between the linear constraint values and
                their lower bounds.
            upper_diffs: The differences between the linear constraint values and
                their upper bounds.

        Returns:
            A tuple containing the transformed lower and upper differences.
        """
        msg = "This transformer does not support linear constraints."
        raise NotImplementedError(msg)


class ObjectiveTransform(ABC):
    """Abstract base class for objective transformations.

    This class defines the interface for transforming objective values between
    the user-defined domain and the optimizer's internal domain. Concrete
    implementations of this class handle the specific logic for each type of
    objective transformation.

    When implementing an objective transformation, the following aspects must be
    considered:

    - **Objective Value Transformation:** Mapping objective values between the
      user and optimizer domains. This is achieved by overriding the
      [`to_optimizer`][ropt.transforms.base.ObjectiveTransform.to_optimizer]
      and
      [`from_optimizer`][ropt.transforms.base.ObjectiveTransform.from_optimizer]
      methods.
    - **Weighted Objective Transformation:** The optimizer works with a
      single, weighted objective value. If the transformation affects the
      weighted objective, the
      [`weighted_objective_from_optimizer`][ropt.transforms.base.ObjectiveTransform.weighted_objective_from_optimizer]
      method should be overridden to handle this.
    """

    @abstractmethod
    def to_optimizer(self, objectives: NDArray[np.float64]) -> NDArray[np.float64]:
        """Transform objective values to the optimizer domain.

        This method maps objective values from the user-defined domain to the
        optimizer's internal domain. This transformation might involve scaling,
        shifting, or other operations to improve the optimizer's performance.

        The input `objectives` may be a multi-dimensional array. It is assumed
        that the last axis of the array represents the objective values. If
        this is not the case, you must adjust the order of the axes before and
        after calling this method.

        Args:
            objectives: The objective values in the user domain to be transformed.

        Returns:
            The transformed objective values in the optimizer domain.
        """

    @abstractmethod
    def from_optimizer(self, objectives: NDArray[np.float64]) -> NDArray[np.float64]:
        """Transform objective values to the user domain.

        This method maps objective values from the optimizer's internal domain
        back to the user-defined domain. This transformation reverses any
        scaling, shifting, or other operations that were applied to improve
        the optimizer's performance.

        The input `objectives` may be a multi-dimensional array. It is assumed
        that the last axis of the array represents the objective values. If
        this is not the case, you must adjust the order of the axes before and
        after calling this method.

        Args:
            objectives: The objective values in the optimizer domain to be transformed.

        Returns:
            The transformed objective values in the user domain.
        """

    def weighted_objective_from_optimizer(
        self, weighted_objective: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Transform the weighted objective to the user domain.

        The optimizer uses a single, weighted objective value evaluated in the
        optimizer domain. This method reverses that transformation, mapping the
        weighted objective back to the user domain.

        For example, if the transformation to the optimizer domain involved a
        sign change to convert a maximization problem into a minimization
        problem, this method would change the sign back.

        Note:
            This method may be applied to the weighted objective itself or to
            its gradient. Therefore, the input may be a scalar or a vector of
            values.

        Args:
            weighted_objective: The weighted objective value(s) to transform.

        Returns:
            The transformed weighted objective value(s).
        """
        return weighted_objective


class NonLinearConstraintTransform(ABC):
    """Abstract base class for nonlinear constraint transformations.

    This class defines the interface for transforming nonlinear constraint
    values between the user-defined domain and the optimizer's internal
    domain. Concrete implementations of this class handle the specific logic
    for each type of nonlinear constraint transformation.

    When implementing a nonlinear constraint transformation, the following
    aspects must be considered:

    - **Constraint Value Transformation:** Mapping constraint values between the
      user and optimizer domains. This is achieved by overriding the
      [`to_optimizer`][ropt.transforms.base.NonLinearConstraintTransform.to_optimizer]
      and
      [`from_optimizer`][ropt.transforms.base.NonLinearConstraintTransform.from_optimizer]
      methods.
    - **Right-Hand-Side Bound Transformation:** Mapping the right-hand-side
      bounds of the constraints between the user and optimizer domains. This is
      achieved by overriding the
      [`bounds_to_optimizer`][ropt.transforms.base.NonLinearConstraintTransform.bounds_to_optimizer]
      method.
    - **Constraint Difference Transformation:** To report violations of
      nonlinear constraints, the differences between constraint values and their
      lower/upper bounds must be transformed from the optimizer domain back to
      the user domain. This is done using the
      [`nonlinear_constraint_diffs_from_optimizer`][ropt.transforms.base.NonLinearConstraintTransform.nonlinear_constraint_diffs_from_optimizer]
      method.
    """

    @abstractmethod
    def to_optimizer(self, constraints: NDArray[np.float64]) -> NDArray[np.float64]:
        """Transform constraint values to the optimizer domain.

        This method maps nonlinear constraint values from the user-defined
        domain to the optimizer's internal domain. This transformation might
        involve scaling, shifting, or other operations to improve the
        optimizer's performance.

        The input `constraints` may be a multi-dimensional array. It is assumed
        that the last axis of the array represents the constraint values. If
        this is not the case, you must adjust the order of the axes before and
        after calling this method.

        Args:
            constraints: The nonlinear constraint values in the user domain to
                be transformed.

        Returns:
            The transformed nonlinear constraint values in the optimizer domain.
        """

    @abstractmethod
    def from_optimizer(self, constraints: NDArray[np.float64]) -> NDArray[np.float64]:
        """Transform constraint values to the user domain.

        This method maps nonlinear constraint values from the optimizer's
        internal domain back to the user-defined domain. This transformation
        reverses any scaling, shifting, or other operations that were applied
        to improve the optimizer's performance.

        The input `constraints` may be a multi-dimensional array. It is assumed
        that the last axis of the array represents the constraint values. If
        this is not the case, you must adjust the order of the axes before and
        after calling this method.

        Args:
            constraints: The nonlinear constraint values in the optimizer domain
                to be transformed.

        Returns:
            The transformed nonlinear constraint values in the user domain.
        """

    @abstractmethod
    def bounds_to_optimizer(
        self, lower_bounds: NDArray[np.float64], upper_bounds: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Transform the right-hand-side bounds to the optimizer domain.

        This method transforms the lower and upper bounds of the nonlinear
        constraints from the user-defined domain to the optimizer's internal
        domain. This transformation is necessary to ensure that the
        constraints remain valid after the variables have been transformed.

        For example, if constraint values are scaled or shifted in the
        optimizer domain, the bounds must be adjusted accordingly.

        Args:
            lower_bounds: The lower bounds on the right-hand-side values in the
                user domain.
            upper_bounds: The upper bounds on the right-hand-side values in the
                user domain.

        Returns:
            A tuple containing the transformed bounds.
        """

    @abstractmethod
    def nonlinear_constraint_diffs_from_optimizer(
        self, lower_diffs: NDArray[np.float64], upper_diffs: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Transform nonlinear constraint differences to the user domain.

        This method transforms the differences between nonlinear constraint
        values and their lower/upper bounds from the optimizer's internal
        domain back to the user-defined domain. These differences are used to
        report constraint violations.

        For example, if constraint values are scaled in the optimizer domain,
        the differences between the constraint values and their bounds must be
        scaled back to the user domain to accurately reflect the constraint
        violations in the user's original units.

        Args:
            lower_diffs: The differences between the nonlinear constraint values
                and their lower bounds.
            upper_diffs: The differences between the nonlinear constraint values
                and their upper bounds.

        Returns:
            A tuple containing the transformed lower and upper differences.
        """



"""This module implements the default optimizer step."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any

import numpy as np
from numpy.typing import NDArray  # noqa: TC002

from ropt.config.enopt import EnOptConfig
from ropt.ensemble_evaluator import EnsembleEvaluator
from ropt.enums import EventType, OptimizerExitCode
from ropt.optimization import EnsembleOptimizer
from ropt.plan import Event, Plan
from ropt.plugins.plan.base import PlanStep
from ropt.results import FunctionResults

if TYPE_CHECKING:
    from numpy.typing import ArrayLike

    from ropt.results import Results
    from ropt.transforms import OptModelTransforms


MetaDataType = dict[str, int | float | bool | str]


class DefaultOptimizerStep(PlanStep):
    """The default optimizer step.

    The optimizer step performs an optimization, yielding a sequence of
    [`FunctionResults`][ropt.results.FunctionResults] and
    [`GradientResults`][ropt.results.GradientResults] objects. The optimizer is
    configured using an [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object or
    a dictionary that can be parsed into such an object. While the initial values
    for optimization are typically specified in the configuration, they can be
    overridden by providing them directly.

    The optimizer step emits several signals:

    - [`START_OPTIMIZER_STEP`][ropt.enums.EventType.START_OPTIMIZER_STEP]:
      Emitted before the optimization starts.
    - [`FINISHED_OPTIMIZER_STEP`][ropt.enums.EventType.FINISHED_OPTIMIZER_STEP]:
      Emitted after the optimization finishes.
    - [`START_EVALUATION`][ropt.enums.EventType.START_EVALUATION]: Emitted
      before a function or gradient evaluation.
    - [`FINISHED_EVALUATION`][ropt.enums.EventType.FINISHED_EVALUATION]: Emitted
      after a function or gradient evaluation.

      The `FINISHED_EVALUATION` signal is particularly important as it passes
      the generated [`Results`][ropt.results.Results] objects. Result handlers
      specified in the plan will respond to this signal to process those results.

    The optimizer step supports nested optimizations, where each function
    evaluation in the optimization calls a function that should run the nested
    optimization and produce the result for the function evaluation.
    """

    def __init__(
        self,
        plan: Plan,
    ) -> None:
        """Initialize a default optimizer step.

        Args:
            plan: The plan that runs this step.
        """
        super().__init__(plan)

    def run(  # type: ignore[override]
        self,
        config: dict[str, Any] | EnOptConfig,
        transforms: OptModelTransforms | None = None,
        variables: ArrayLike | None = None,
        nested_optimization: Plan | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> OptimizerExitCode:
        """Run the optimizer step.

        Args:
            config:              The optimizer configuration.
            transforms:          Optional transforms object.
            variables: Variables to evaluate.
            nested_optimization: Optional nested plan.
            metadata:            Optional metadata to add to events.
        """
        self._config = EnOptConfig.model_validate(config, context=transforms)
        self._transforms = transforms
        self._nested_optimization = nested_optimization
        self._metadata = metadata

        self.emit_event(
            Event(
                event_type=EventType.START_OPTIMIZER_STEP,
                config=self._config,
                source=self.id,
            )
        )

        if variables is None:
            variables = self._config.variables.initial_values
        else:
            variables = np.array(np.asarray(variables, dtype=np.float64), ndmin=1)
            if transforms is not None and transforms.variables is not None:
                variables = transforms.variables.to_optimizer(variables)

        ensemble_evaluator = EnsembleEvaluator(
            self._config,
            self._transforms,
            self.plan.optimizer_context.evaluator,
            self.plan.optimizer_context.plugin_manager,
        )

        ensemble_optimizer = EnsembleOptimizer(
            enopt_config=self._config,
            ensemble_evaluator=ensemble_evaluator,
            plugin_manager=self.plan.optimizer_context.plugin_manager,
            nested_optimizer=(
                self._run_nested_plan if self._nested_optimization is not None else None
            ),
            signal_evaluation=self._signal_evaluation,
        )

        if ensemble_optimizer.is_parallel and self._nested_optimization is not None:
            msg = "Nested optimization detected: parallel evaluation not supported. "
            raise RuntimeError(msg)

        exit_code = ensemble_optimizer.start(variables)

        if exit_code == OptimizerExitCode.USER_ABORT:
            self.plan.abort()

        self.emit_event(
            Event(
                event_type=EventType.FINISHED_OPTIMIZER_STEP,
                config=self._config,
                source=self.id,
            )
        )

        return exit_code

    def emit_event(self, event: Event) -> None:
        """Emit an event.

        Args:
            event: The event to emit.
        """
        self.plan.emit_event(event)

    def _signal_evaluation(self, results: tuple[Results, ...] | None = None) -> None:
        if results is None:
            self.emit_event(
                Event(
                    event_type=EventType.START_EVALUATION,
                    config=self._config,
                    source=self.id,
                )
            )
        else:
            if self._metadata is not None:
                for item in results:
                    item.metadata = deepcopy(self._metadata)

            data: dict[str, Any] = {"results": results}
            self.emit_event(
                Event(
                    event_type=EventType.FINISHED_EVALUATION,
                    config=self._config,
                    source=self.id,
                    data=data,
                ),
            )

    def _run_nested_plan(
        self, variables: NDArray[np.float64]
    ) -> tuple[FunctionResults | None, bool]:
        if self._nested_optimization is None:
            return None, False
        self._nested_optimization.set_parent(self.plan)
        results = self._nested_optimization.run_function(variables)
        if self._nested_optimization.aborted:
            self.plan.abort()
        if not isinstance(results, FunctionResults):
            msg = "Nested optimization must return a FunctionResults object."
            raise TypeError(msg)
        return results, self._nested_optimization.aborted



"""This module defines the abstract base class for plugins."""

from __future__ import annotations

from abc import ABC, abstractmethod


class Plugin(ABC):
    """Abstract base class for plugins.

    `ropt`  plugins should derive from this base class, which specifies an
    `is_supported` method.
    """

    @classmethod
    @abstractmethod
    def is_supported(cls, method: str) -> bool:
        """Check whether a given method is supported.

        This method is called by the
        [`is_supported`][ropt.plugins.PluginManager.is_supported] method of
        [`PluginManager`][ropt.plugins.PluginManager] objects to verify if a
        specific method is supported by this plugin.

        Args:
            method:   The name of the method to check.

        Returns:
            True if the method is supported; otherwise, False.
        """

    @classmethod
    def allows_discovery(cls) -> bool:
        """Check if the plugin can be discovered automatically.

        Normally, plugins may be discovered automatically by the plugin manager
        by checking if they support a specific method. However, some plugins may
        not support this behavior and should be explicitly requested. These
        should override this property to return `False`.

        For example, the
        [`external`][ropt.plugins.optimizer.external.ExternalOptimizer]
        optimizer plugin does not define its own methods but launches methods
        from other plugins as an external process. Therefore, the `external`
        optimizer plugin must always be specified explicitly, and this method
        is overloaded to return False.
        """
        return True



"""Configuration class for realizations."""

from __future__ import annotations

from typing import Self

import numpy as np
from pydantic import ConfigDict, NonNegativeInt, model_validator

from ropt.config.utils import ImmutableBaseModel, normalize
from ropt.config.validated_types import Array1D  # noqa: TC001


class RealizationsConfig(ImmutableBaseModel):
    """Configuration class for realizations.

    This class, `RealizationsConfig`, defines the configuration for realizations
    used in an [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object.

    To optimize an ensemble of functions, a set of realizations is defined. When
    the optimizer requests a function value or a gradient, these are calculated for
    each realization and then combined into a single value. Typically, this
    combination is a weighted sum, but other methods are possible.

    The `weights` field, a `numpy` array, determines the weight of each
    realization. The length of this array defines the number of realizations. The
    weights are automatically normalized to sum to 1 (e.g., `[1, 1]` becomes
    `[0.5, 0.5]`).

    If function value calculations for some realizations fail (e.g., due to a
    simulation error), the total function and gradient values can still be
    calculated by excluding the missing values. However, a minimum number of
    successful realizations may be required. The `realization_min_success` field
    specifies this minimum. By default, it is set equal to the number of
    realizations, meaning no missing values are allowed.

    Note:
        Setting `realization_min_success` to zero allows the optimization to
        proceed even if all realizations fail. While some optimizers can handle
        this, most will treat it as if the value were one, requiring at least
        one successful realization.

    Attributes:
        weights:                 Weights for the realizations (default: 1.0).
        realization_min_success: Minimum number of successful realizations (default:
                                equal to the number of realizations).
    """

    weights: Array1D = np.array(1.0)
    realization_min_success: NonNegativeInt | None = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        validate_default=True,
    )

    @model_validator(mode="after")
    def _broadcast_normalize_and_check(self) -> Self:
        self._mutable()
        self.weights = normalize(self.weights)
        if (
            self.realization_min_success is None
            or self.realization_min_success > self.weights.size
        ):
            self.realization_min_success = self.weights.size
        self._immutable()
        return self



from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any

import numpy as np
import pytest
from numpy.random import Generator, default_rng

from ropt.config.enopt import EnOptConfig
from ropt.ensemble_evaluator._gradient import _perturb_variables
from ropt.plan import BasicOptimizer
from ropt.plugins.sampler.base import Sampler, SamplerPlugin

if TYPE_CHECKING:
    from numpy.typing import NDArray


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> dict[str, Any]:
    return {
        "optimizer": {
            "tolerance": 0.005,
            "max_functions": 4,
        },
        "objectives": {
            "weights": [0.75, 0.24],
        },
        "samplers": [
            {
                "method": "test",
            },
        ],
        "gradient": {
            "number_of_perturbations": 3,
            "perturbation_magnitudes": 0.01,
        },
        "variables": {
            "initial_values": [0, 0, 0],
            "upper_bounds": 1.0,
            "lower_bounds": -1.0,
        },
    }


class MockedSampler(Sampler):
    def __init__(
        self,
        enopt_config: EnOptConfig,
        sampler_index: int,
        mask: NDArray[np.bool_] | None,
        _: Generator,
    ) -> None:
        self._config = enopt_config
        self._sampler = enopt_config.samplers[sampler_index]
        self._mask = mask
        # This sampler only works if the number of perturbation equals the
        # number of variables:
        assert enopt_config.gradient.number_of_perturbations == len(
            enopt_config.variables.initial_values
        )

    def generate_samples(self) -> NDArray[np.float64]:
        variable_count = self._config.variables.initial_values.size
        realization_count = self._config.realizations.weights.size
        perturbation_count = self._config.gradient.number_of_perturbations

        samples: NDArray[np.float64]
        if self._mask is None:
            samples = np.ones(
                (realization_count, perturbation_count, variable_count),
                dtype=np.float64,
            )
        else:
            samples = np.zeros(
                (realization_count, perturbation_count, variable_count),
                dtype=np.float64,
            )
            samples[..., self._mask] = 1.0
        if "scale" in self._sampler.options:
            samples *= self._sampler.options["scale"]
        for idx in range(samples.shape[0]):
            diag = np.diag(samples[idx, ...])
            samples[idx, ...] = np.diag(diag)
        return samples


class MockedSamplerPlugin(SamplerPlugin):
    @classmethod
    def create(
        cls,
        enopt_config: EnOptConfig,
        sampler_index: int,
        mask: NDArray[np.bool_] | None,
        rng: Generator,
    ) -> MockedSampler:
        return MockedSampler(enopt_config, sampler_index, mask, rng)

    @classmethod
    def is_supported(cls, method: str) -> bool:
        return method.lower() in {"test"}


def test_sampler_simple(enopt_config: Any) -> None:
    enopt_config["gradient"]["number_of_perturbations"] = 3
    rng = default_rng(123)
    config = EnOptConfig.model_validate(enopt_config)
    sampler = MockedSampler(config, 0, None, rng)

    perturbed_variables = _perturb_variables(
        config,
        np.array([0.0, 0.0, 0.0]),
        [sampler],
    )
    assert np.allclose(perturbed_variables, 0.01 * np.eye(3))


def test_sampler_use_options(enopt_config: Any) -> None:
    rng = default_rng(123)
    samplers: list[dict[str, Any]] = enopt_config["samplers"]
    samplers[0]["options"] = {"scale": 100.0}
    config = EnOptConfig.model_validate(enopt_config)
    sampler = MockedSampler(config, 0, None, rng)

    perturbed_variables = _perturb_variables(
        config,
        np.array([0.0, 0.0, 0.0]),
        [sampler],
    )
    assert np.allclose(perturbed_variables, np.eye(3))


def test_sampler_indexed(enopt_config: Any) -> None:
    rng = default_rng(123)
    samplers: list[dict[str, Any]] = enopt_config["samplers"]
    samplers.append(copy.deepcopy(samplers[0]))
    samplers[1]["options"] = {"scale": -1}
    enopt_config["gradient"]["samplers"] = [0, 1, 1]
    config = EnOptConfig.model_validate(enopt_config)
    sampler1 = MockedSampler(config, 0, np.array([0]), rng)
    sampler2 = MockedSampler(config, 1, np.array([1, 2]), rng)

    perturbed_variables = _perturb_variables(
        config,
        np.array([0.0, 0.0, 0.0]),
        [sampler1, sampler2],
    )
    assert np.allclose(perturbed_variables[0, ...], np.diag([0.01, -0.01, -0.01]))


def test_sampler_order(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["gradient"]["samplers"] = [0, 0, 1]
    enopt_config["samplers"] = [
        {"method": "norm"},
        {"method": "uniform"},
    ]
    variables1 = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables1 is not None
    assert np.allclose(variables1, [0, 0, 0.5], atol=0.025)

    # Switch the samplers:
    enopt_config["samplers"] = [
        {"method": "uniform"},
        {"method": "norm"},
    ]
    enopt_config["gradient"]["samplers"] = [1, 1, 0]
    variables2 = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables2 is not None
    assert np.allclose(variables2, [0, 0, 0.5], atol=0.025)

    assert np.allclose(variables1, variables2)



"""This module implements the default tracker handler."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from ropt.enums import EventType
from ropt.plugins.plan.base import ResultHandler

from ._utils import _get_last_result, _update_optimal_result

if TYPE_CHECKING:
    import uuid

    from ropt.plan import Event, Plan
    from ropt.results import FunctionResults
    from ropt.transforms import OptModelTransforms


class DefaultTrackerHandler(ResultHandler):
    """The default tracker results handler object.

    This handler tracks the
    [`Results`][ropt.results.Results] objects that it receives and selects one
    to retain in a variable. Currently it tracks either the last result it
    receives, or the best result. The best result is defined as the result that
    has the lowest weighted objective value. Optionally, results may be filtered
    by checking for violations of constraints, by comparing constraint values to
    a threshold.
    """

    def __init__(
        self,
        plan: Plan,
        *,
        what: Literal["best", "last"] = "best",
        constraint_tolerance: float | None = None,
        sources: set[uuid.UUID] | None = None,
        transforms: OptModelTransforms | None = None,
    ) -> None:
        """Initialize a default tracker results handler.

        This handler monitors `Results` objects from specified sources and
        selects a single result to retain based on the specified criteria. It
        can track either the best result encountered so far or the most recent
        result. The "best" result is determined by the lowest weighted
        objective value.

        Results can optionally be filtered based on constraint violations. If a
        `constraint_tolerance` is provided, results that violate constraints
        beyond this tolerance will be discarded and not tracked.

        The `sources` parameter allows you to specify which steps' results
        should be tracked. Only results from steps whose IDs are included in
        this set will be considered.

        Tracking of results is based on the results in the optimizer domain, and
        the tracked results are accessible via the `"results"` key. If the
        `transforms` argument is not `None`, tracking still occurs in the
        optimizer domain, but the results accessible via the `"results"` key are
        transformed to the user domain.

        Args:
            plan:                 The plan this handler is part of.
            what:                 Specifies whether to track the "best" or "last" result.
            constraint_tolerance: Constraint tolerance for filtering results.
            sources:              A set of UUIDs of steps whose results to track.
            transforms:           Optional transforms to apply to the final results.
        """
        super().__init__(plan)
        self._what = what
        self._constraint_tolerance = constraint_tolerance
        self._sources = set() if sources is None else sources
        self._transforms = transforms
        self._tracked_results: FunctionResults | None = None
        self["results"] = None

    def handle_event(self, event: Event) -> None:
        """Handle an event.

        Args:
            event: The event to handle.
        """
        if (
            event.event_type == EventType.FINISHED_EVALUATION
            and "results" in event.data
            and event.source in self._sources
        ):
            results = event.data["results"]
            if self["results"] is None:
                self._tracked_results = None
            filtered_results: FunctionResults | None = None
            match self._what:
                case "best":
                    filtered_results = _update_optimal_result(
                        self._tracked_results,
                        results,
                        self._constraint_tolerance,
                    )
                case "last":
                    filtered_results = _get_last_result(
                        results,
                        self._constraint_tolerance,
                    )
            if filtered_results is not None:
                self._tracked_results = filtered_results
                if self._transforms is not None:
                    filtered_results = filtered_results.transform_from_optimizer(
                        self._transforms
                    )
                self["results"] = filtered_results



"""This module defines the base classes for optimization plan plugins.

Optimization plan steps and result handlers can be added via the plugin
mechanism to implement additional functionality. This is done by creating plugin
classes that derive from the
[`PlanHandlerPlugin`][ropt.plugins.plan.base.PlanHandlerPlugin] and
[`PlanStepPlugin`][ropt.plugins.plan.base.PlanStepPlugin] classes.
"""

from __future__ import annotations

import uuid
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from ropt.plugins.base import Plugin

if TYPE_CHECKING:
    from ropt.plan import Event, Plan


class PlanHandlerPlugin(Plugin):
    """Abstract base class for plan handler plugins.

    This base class serves as the foundation for handler plugins used in an
    optimization plan. Any plugin derived from this class can be built-in,
    installed via a plugin mechanism, or dynamically loaded. During the
    execution of an optimization plan, the required plugin is located through
    the [`PluginManager`][ropt.plugins.PluginManager], which then uses the
    plugin's `create_*` functions to instantiate  a
    [`ResultHandler`][ropt.plugins.plan.base.ResultHandler].
    """

    @classmethod
    @abstractmethod
    def create(
        cls,
        name: str,
        plan: Plan,
        **kwargs: Any,  # noqa: ANN401
    ) -> ResultHandler:
        """Create a result handler.

        This factory function instantiates result handler
        object based on the provided configuration.

        Args:
            name:   The name of the handler.
            plan:   The plan in which the handler operates.
            kwargs: Additional arguments to pass to the handler.
        """


class PlanStepPlugin(Plugin):
    """Abstract base class for plan step plugins.

    This base class serves as the foundation for step plugins used in an
    optimization plan. Any plugin derived from this class can be built-in,
    installed via a plugin mechanism, or dynamically loaded. During the
    execution of an optimization plan, the required plugin is located through
    the [`PluginManager`][ropt.plugins.PluginManager], which then uses the
    plugin's `create_*` functions to instantiate a
    [`PlanStep`][ropt.plugins.plan.base.PlanStep].
    """

    @classmethod
    @abstractmethod
    def create(
        cls,
        name: str,
        plan: Plan,
        **kwargs: Any,  # noqa: ANN401
    ) -> PlanStep:
        """Create a result handler.

        This factory function instantiates result handler
        object based on the provided configuration.

        Args:
            name: The name of the step.
            plan: The plan in which the step operates.
            kwargs: Additional arguments to pass to the step.
        """


class PlanStep(ABC):
    """Base class for steps.

    steps loaded and created by the plugin manager must inherit from this
    abstract base class. It defines a `run` method that must be overridden to
    implement the specific functionality of each step.
    """

    def __init__(self, plan: Plan) -> None:
        """Initialize the step.

        The `plan` argument is accessible as the `plan` property.

        Args:
            plan:   The parent plan that manages this step.
        """
        self.__stored_plan = plan
        self.__id: uuid.UUID = uuid.uuid4()

    @property
    def id(self) -> uuid.UUID:
        """Return the unique identifier of the handler.

        Each handler instance is assigned a unique ID upon creation. This ID
        can be used to distinguish between different handler instances.

        Returns:
            A UUID object representing the unique identifier of the handler.
        """
        return self.__id

    @property
    def plan(self) -> Plan:
        """Return the plan that executes the step.

        Returns:
            The plan object.
        """
        return self.__stored_plan

    @abstractmethod
    def run(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Execute the step object.

        This method must be overloaded to implement the functionality of the
        step.

        Args:
            args:   Optional positional arguments.
            kwargs: Optional keyword arguments.
        """


class ResultHandler(ABC):
    """Base class for result handler objects.

    Result handlers loaded and created by the plugin manager must inherit from
    this abstract base class. It defines a `handle_event` method that must be
    overridden to implement specific functionality.
    """

    def __init__(self, plan: Plan) -> None:
        """Initialize a results handler.

        The `plan` argument is accessible as the `plan` property.

        Args:
            plan:   The parent plan that contains the object.
        """
        self.__stored_plan = plan
        self.__stored_values: dict[str, Any] = {}
        self.__id: uuid.UUID = uuid.uuid4()

    @property
    def id(self) -> uuid.UUID:
        """Return the unique identifier of the handler.

        Each handler instance is assigned a unique ID upon creation. This ID
        can be used to distinguish between different handler instances.

        Returns:
            A UUID object representing the unique identifier of the handler.
        """
        return self.__id

    @property
    def plan(self) -> Plan:
        """Return the associated plan that executes the handler.

        Returns:
            The plan object.
        """
        return self.__stored_plan

    @abstractmethod
    def handle_event(self, event: Event) -> None:
        """Handle and propagate an event.

        This method must be overloaded to implement the functionality of the
        handler, based on the information passed via the event object.
        """

    def __getitem__(self, key: str) -> Any:  # noqa: ANN401
        """Get the value of a plan variable.

        This method implements the `[]` operator on the handler object to retrieve
        the value associated with a specific key.

        Args:
            key: The key to retrieve.

        Returns:
            The value corresponding to the key.
        """
        if key in self.__stored_values:
            return self.__stored_values[key]
        msg = f"Unknown plan variable: `{key}`"
        raise AttributeError(msg)

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: ANN401
        """Set a plan variable to the given value.

        This method implements the `[]` operator on the handler object to store
        arbitrary values.

        Args:
            key:   The key to set.
            value: The value to store.
        """
        if not key.isidentifier():
            msg = f"Not a valid variable name: `{key}`"
            raise AttributeError(msg)
        self.__stored_values[key] = value



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from importlib.util import find_spec
from typing import TYPE_CHECKING, Any, Final, Iterable, Sequence, TypeVar

if TYPE_CHECKING:
    from ropt.enums import ResultAxis
    from ropt.transforms._transforms import OptModelTransforms


_HAVE_PANDAS: Final = find_spec("pandas") is not None

if TYPE_CHECKING and _HAVE_PANDAS:
    import pandas as pd  # noqa: TC002
if _HAVE_PANDAS:
    from ._pandas import _to_dataframe

TypeResults = TypeVar("TypeResults", bound="Results")


@dataclass(slots=True)
class Results(ABC):
    """Abstract base class for storing optimization results.

    The `Results` class serves as a foundation for storing various types of
    optimization results. It is not intended to be instantiated directly but
    rather serves as a base for derived classes like
    [`FunctionResults`][ropt.results.FunctionResults] and
    [`GradientResults`][ropt.results.GradientResults], which hold the actual
    data.

    This class provides storage for the following generic information:

    *   **Batch ID:** An optional identifier, potentially generated by the
        function evaluator, that uniquely identifies a group of function
        evaluations passed to the evaluator by teh optimizer.
    *   **Metadata:** A dictionary for storing additional information generated
        during optimization. This metadata can include various primitive values
        that are not directly interpreted by the optimization code but are
        useful for reporting and analysis.

    The derived classes, [`FunctionResults`][ropt.results.FunctionResults] and
    [`GradientResults`][ropt.results.GradientResults], extend this base class
    with specific attributes for storing function evaluation results and
    gradient evaluation results, respectively. These derived classes also
    provide methods for exporting the stored data.

    One key method provided by the `Results` class is
    [`to_dataframe`][ropt.results.Results.to_dataframe], which allows exporting
    the contents of a specific field, or a subset of its sub-fields, to a
    `pandas` DataFrame for further data analysis and reporting.

    Attributes:
        batch_id: The ID of the evaluation batch.
        metadata: A dictionary of metadata.
    """

    batch_id: int | None
    metadata: dict[str, Any]

    def to_dataframe(
        self,
        field_name: str,
        select: Iterable[str],
        unstack: Iterable[ResultAxis] | None = None,
        names: dict[str, Sequence[str | int] | None] | None = None,
    ) -> pd.DataFrame:
        """Export a field to a pandas DataFrame.

        Exports the values of a single field to a `pandas` DataFrame.
        The field to export is selected by the `field_name` argument.
        Typically, such a field contains multiple sub-fields. By default,
        all sub-fields are exported as columns in the DataFrame, but a
        subset can be selected using the `select` argument.

        Sub-fields may be multi-dimensional arrays, which are exported in a
        stacked manner. Using the axis types found in the metadata, the exporter
        constructs a multi-index labeled with the corresponding names provided
        via the `names` argument. If `names` is `None`, numerical indices are
        used. These multi-indices can optionally be unstacked into multiple
        columns by providing the axis types to unstack via the `unstack`
        argument.

        Info: The DataFrame Index
            The index of the resulting DataFrame may be a multi-index
            constructed from axis indices or labels. In addition, the `batch_id`
            (if not `None`) is prepended to the index.

        Args:
            field_name: The field to export.
            select:     Select the sub-fields to export. By default, all
                        sub-fields are exported.
            unstack:    Select axes to unstack. By default, no axes are
                        unstacked.
            names:      A dictionary mapping axis types to names.

        Raises:
            NotImplementedError: If the `pandas` module is not installed.
            ValueError:          If the field name is incorrect.

        Returns:
            A `pandas` DataFrame containing the results.

        Warning:
            This function requires the `pandas` module to be installed.
        """
        if not _HAVE_PANDAS:
            msg = "The pandas module must be installed to use to_dataframe"
            raise NotImplementedError(msg)

        result_field = getattr(self, field_name, None)
        if result_field is None:
            msg = f"Invalid result field: {field_name}"
            raise AttributeError(msg)

        return _to_dataframe(
            result_field,
            self.batch_id,
            select,
            unstack,
            names,
        )

    @abstractmethod
    def transform_from_optimizer(self, transforms: OptModelTransforms) -> Results:
        """Transform results from the optimizer domain to the user domain.

        During optimization, variables, objectives, and constraints are often
        transformed to a different domain (the optimizer domain) to enhance
        the performance and stability of the optimization algorithm. The
        [`Results`][ropt.results.Results] objects produced during optimization
        are initially in the optimizer domain. This method reverses these
        transformations, mapping the results back to the user-defined domain.
        The transformations between the user and optimizer domains are defined
        by the classes in the [`ropt.transforms`][ropt.transforms] module.

        For instance, variables might have been scaled and shifted to a range
        more suitable for the optimizer. This method, using the provided
        `OptModelTransforms` object, applies the inverse scaling and shifting to
        restore the variables to their original scale and offset. Similarly,
        objectives and constraints are transformed back to the user domain.

        These transformations are defined and managed by the
        [`OptModelTransforms`][ropt.transforms.OptModelTransforms] object,
        which encapsulates the specific transformations for variables,
        objectives, and nonlinear constraints.

        Args:
            transforms: The `OptModelTransforms` object containing the domain
                transformations to apply.

        Returns:
            A new `FunctionResults` object with all relevant data transformed
            back to the user domain.
        """



"""Generate a results report in a `pandas` data frame."""

from __future__ import annotations

from functools import partial
from importlib.util import find_spec
from typing import TYPE_CHECKING, Any, Final, Literal, Sequence

from ropt.enums import ResultAxis

from ._function_results import FunctionResults
from ._gradient_results import GradientResults

if TYPE_CHECKING:
    from ropt.results import Results

_HAVE_PANDAS: Final = find_spec("pandas") is not None

if TYPE_CHECKING:
    from ropt.results import Results

if _HAVE_PANDAS:
    import pandas as pd


def _get_function_results(
    results: Results,
    sub_fields: set[str],
    names: dict[str, Sequence[str | int] | None] | None,
) -> pd.DataFrame:
    if (
        not sub_fields
        or not isinstance(results, FunctionResults)
        or results.functions is None
    ):
        return pd.DataFrame()

    functions = results.to_dataframe(
        "functions",
        select=_get_select("functions", sub_fields),
        unstack=[ResultAxis.OBJECTIVE, ResultAxis.NONLINEAR_CONSTRAINT],
        names=names,
    ).rename(columns=partial(_add_prefix, prefix="functions"))

    evaluations = results.to_dataframe(
        "evaluations",
        select=_get_select("evaluations", sub_fields),
        unstack=[
            ResultAxis.VARIABLE,
            ResultAxis.OBJECTIVE,
            ResultAxis.NONLINEAR_CONSTRAINT,
        ],
        names=names,
    ).rename(columns=partial(_add_prefix, prefix="evaluations"))

    if results.constraint_info is not None:
        constraint_info = results.to_dataframe(
            "constraint_info",
            select=_get_select("constraint_info", sub_fields),
            unstack=[
                ResultAxis.VARIABLE,
                ResultAxis.LINEAR_CONSTRAINT,
                ResultAxis.NONLINEAR_CONSTRAINT,
            ],
            names=names,
        ).rename(columns=partial(_add_prefix, prefix="constraint_info"))

        return _join_frames(functions, evaluations, constraint_info)

    return _join_frames(functions, evaluations)


def _get_gradient_results(
    results: Results,
    sub_fields: set[str],
    names: dict[str, Sequence[str | int] | None] | None,
) -> pd.DataFrame:
    if (
        not sub_fields
        or not isinstance(results, GradientResults)
        or results.gradients is None
    ):
        return pd.DataFrame()

    gradients = results.to_dataframe(
        "gradients",
        select=_get_select("gradients", sub_fields),
        unstack=[
            ResultAxis.OBJECTIVE,
            ResultAxis.NONLINEAR_CONSTRAINT,
            ResultAxis.VARIABLE,
        ],
        names=names,
    ).rename(columns=partial(_add_prefix, prefix="gradients"))

    evaluations = results.to_dataframe(
        "evaluations",
        select=_get_select("evaluations", sub_fields),
        unstack=[
            ResultAxis.VARIABLE,
            ResultAxis.OBJECTIVE,
            ResultAxis.NONLINEAR_CONSTRAINT,
        ],
        names=names,
    ).rename(columns=partial(_add_prefix, prefix="evaluations"))

    return _join_frames(gradients, evaluations)


def _join_frames(*args: pd.DataFrame) -> pd.DataFrame:
    frames = [frame for frame in args if not frame.empty]
    return (
        frames[0].join(list(frames[1:]), how="outer") if len(frames) > 1 else frames[0]
    )


def _get_select(field_name: str, sub_fields: set[str]) -> list[str]:
    return [
        item.removeprefix(f"{field_name}.")
        for item in sub_fields
        if item.startswith(f"{field_name}.")
    ]


def _add_prefix(name: tuple[str, ...] | str, prefix: str) -> tuple[str, ...] | str:
    return (
        (f"{prefix}.{name[0]}",) + name[1:]
        if isinstance(name, tuple)
        else f"{prefix}.{name}"
    )


def _add_metadata(
    data_frame: pd.DataFrame, results: Results, sub_fields: set[str]
) -> pd.DataFrame:
    for field in sub_fields:
        split_fields = field.split(".")
        if split_fields[0] == "metadata":
            value = _get_value(results.metadata, split_fields[1:])
            if value is not None:
                data_frame[field] = value
    return data_frame


def _get_value(data: dict[str, Any], keys: list[str]) -> Any | None:  # noqa: ANN401
    for key in keys:
        if isinstance(data, dict):
            if key not in data:
                return None
            data = data[key]
        else:
            break
    return data


def results_to_dataframe(
    results: Sequence[Results],
    fields: set[str],
    result_type: Literal["functions", "gradients"],
    names: dict[str, Sequence[str | int] | None] | None = None,
) -> pd.DataFrame:
    """Combine a sequence of results into a single pandas DataFrame.

    This function aggregates results from multiple
    [`FunctionResults`][ropt.results.FunctionResults] or
    [`GradientResults`][ropt.results.GradientResults] objects into a single
    `pandas` DataFrame. It is designed to be used with observers that produce
    results during the optimization process.

    The `fields` argument determines which data fields to include in the
    DataFrame. These fields can be any of the attributes defined within
    [`FunctionResults`][ropt.results.FunctionResults] or
    [`GradientResults`][ropt.results.GradientResults]. Nested fields are
    specified using dot notation (e.g., `evaluations.variables` to include the
    `variables` field within the `evaluations` object).

    The `evaluation_info` sub-fields, found within the `evaluations` fields of
    [`functions`][ropt.results.FunctionEvaluations] and
    [`gradient`][ropt.results.GradientEvaluations] results, respectively, are
    dictionaries. To include specific keys from these dictionaries, use the
    format `evaluations.evaluation_info.key`, where `key` is the name of the
    desired key.

    Many fields may result in multiple columns in the DataFrame. For example,
    `evaluations.variables` will generate a separate column for each variable.
    If available, variable names will be used as column labels.
    Multi-dimensional fields, such as those with named realizations and
    objectives, will have column names that are tuples of the corresponding
    names.

    The `result_type` argument specifies whether to include function evaluation
    results (`functions`) or gradient results (`gradients`).

    The `names` argument is an optional dictionary that maps axis types to
    names. These names are used to label the multi-index columns in the
    resulting DataFrame. If not provided, numerical indices are used.

    Args:
        results:     A sequence of [`Results`][ropt.results.Results] objects
                     to combine.
        fields:      The names of the fields to include in the DataFrame.
        result_type: The type of results to include ("functions" or
                     "gradients").
        names:       A dictionary mapping axis types to names.

    Returns:
        A `pandas` DataFrame containing the combined results.
    """
    if result_type not in ("functions", "gradients"):
        msg = f"Invalid frame output type: {result_type}"
        raise TypeError(msg)

    frame = pd.DataFrame()
    for item in results:
        if not isinstance(item, (FunctionResults, GradientResults)):
            msg = f"Invalid result type: {type(item)}"
            raise TypeError(msg)

        if (
            result_type == "functions"
            and isinstance(item, FunctionResults)
            and item.functions is not None
        ):
            frame = pd.concat(
                [
                    frame,
                    _add_metadata(
                        _get_function_results(item, fields, names), item, fields
                    ),
                ]
            )
        elif (
            result_type == "gradients"
            and isinstance(item, GradientResults)
            and item.gradients is not None
        ):
            frame = pd.concat(
                [
                    frame,
                    _add_metadata(
                        _get_gradient_results(item, fields, names), item, fields
                    ),
                ]
            )

    return frame



import numpy as np
from numpy.typing import NDArray

from ropt.config.enopt import EnOptConfig
from ropt.enums import BoundaryType
from ropt.plugins.function_estimator.base import FunctionEstimator
from ropt.plugins.sampler.base import Sampler

SVD_TOLERANCE = 0.999
MIRROR_REPEAT = 3


def _apply_bounds(
    variables: NDArray[np.float64],
    lower_bounds: NDArray[np.float64],
    upper_bounds: NDArray[np.float64],
    truncation_types: NDArray[np.ubyte],
) -> NDArray[np.float64]:
    def mirror(
        variables: NDArray[np.float64],
        mask: NDArray[np.bool_],
        condition: NDArray[np.bool_],
        bounds: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        return np.where(
            np.logical_and(mask, condition), 2 * bounds - variables, variables
        )

    # Repeat the mirroring a few times, handling mirrored values that still
    # violate the bounds. If that is not sufficient, clip the values.
    mask1 = np.logical_and(
        truncation_types == BoundaryType.MIRROR_BOTH, variables < lower_bounds
    )
    mask2 = np.logical_and(
        truncation_types == BoundaryType.MIRROR_BOTH, variables > upper_bounds
    )
    for _ in range(MIRROR_REPEAT):
        variables = mirror(variables, mask1, variables < lower_bounds, lower_bounds)
        variables = mirror(variables, mask1, variables > upper_bounds, upper_bounds)
    for _ in range(MIRROR_REPEAT):
        variables = mirror(variables, mask2, variables > upper_bounds, upper_bounds)
        variables = mirror(variables, mask2, variables < lower_bounds, lower_bounds)

    # Finally, fall back to clipping.
    return np.clip(variables, lower_bounds, upper_bounds)


def _invert_linear_equations(
    matrix: NDArray[np.float64], vector: NDArray[np.float64]
) -> NDArray[np.float64]:
    u, sigma, v = np.linalg.svd(matrix)
    u = u[:, : sigma.size]
    v = v[: sigma.size, :]
    sigma2 = sigma**2
    select = np.cumsum(sigma2) / np.sum(sigma2) < SVD_TOLERANCE
    select[np.argmin(select)] = True  # Add the element that passes the tolerance
    sigma_inv = np.diag(
        np.divide(1.0, sigma, out=np.zeros_like(sigma), where=(sigma > 0) & select)
    )
    result: NDArray[np.float64] = v.T.dot(sigma_inv).dot(u.T).dot(vector)
    return result


def _perturb_variables(
    config: EnOptConfig,
    variables: NDArray[np.float64],
    samplers: list[Sampler],
) -> NDArray[np.float64]:
    if config.gradient.samplers is None:
        samples = samplers[0].generate_samples()
    else:
        # The results should be independent of the order of the samplers,
        # reordering would affect the random numbers they are based on. We
        # obtain a consistent order by running multiple samplers in the order
        # that they appear in the config.gradient.samplers array:
        unique, indices = np.unique(
            np.compress(
                config.gradient.samplers >= 0,
                config.gradient.samplers,
            ),
            return_index=True,
        )
        sampler_indices = unique[np.argsort(indices)]
        samples = samplers[sampler_indices[0]].generate_samples()
        for sampler_idx in sampler_indices[1:]:
            samples += samplers[sampler_idx].generate_samples()
    return _apply_bounds(
        variables + config.gradient.perturbation_magnitudes * samples,
        config.variables.lower_bounds,
        config.variables.upper_bounds,
        config.gradient.boundary_types,
    )


def _estimate_gradients(
    delta_variables: NDArray[np.float64],
    delta_functions: NDArray[np.float64],
    weights: NDArray[np.float64],
) -> NDArray[np.float64]:
    active_realizations = np.abs(weights) > 0
    realization_count = active_realizations.size
    gradients: NDArray[np.float64] = np.zeros(
        (delta_variables.shape[-1], realization_count), dtype=np.float64
    )
    all_successes = np.logical_not(np.isnan(delta_functions))
    for idx in range(realization_count):
        success = all_successes[idx]
        if active_realizations[idx] and np.any(success):
            ensemble_matrix = delta_variables[idx, success, :]
            gradients[:, idx] = _invert_linear_equations(
                ensemble_matrix, delta_functions[idx, success]
            )
    return gradients


def _estimate_merged_gradient(
    delta_variables: NDArray[np.float64],
    delta_functions: NDArray[np.float64],
    weights: NDArray[np.float64],
) -> NDArray[np.float64]:
    delta_functions = delta_functions * np.expand_dims(weights, axis=-1)
    active_realizations = np.abs(weights) > 0
    active_perturbations = np.repeat(active_realizations, delta_variables.shape[1])
    delta_variables = delta_variables.reshape(-1, delta_variables.shape[-1])
    delta_functions = delta_functions.flatten()
    active_perturbations &= np.logical_not(np.isnan(delta_functions))
    return _invert_linear_equations(
        delta_variables[active_perturbations, :], delta_functions[active_perturbations]
    )


def _calculate_gradient(  # noqa: PLR0913
    functions: NDArray[np.float64],
    delta_variables: NDArray[np.float64],
    delta_functions: NDArray[np.float64],
    failed_realizations: NDArray[np.bool_],
    weights: NDArray[np.float64],
    estimator: FunctionEstimator,
    *,
    merge_realizations: bool,
) -> NDArray[np.float64]:
    weights = np.where(failed_realizations, 0.0, weights)
    weights /= weights.sum()
    if merge_realizations:
        gradients = _estimate_merged_gradient(delta_variables, delta_functions, weights)
    else:
        gradients = _estimate_gradients(delta_variables, delta_functions, weights)
    return estimator.calculate_gradient(functions, gradients, weights)


def _calculate_estimated_gradients(  # noqa: PLR0913
    config: EnOptConfig,
    function_estimators: list[FunctionEstimator],
    estimator_indices: NDArray[np.intc] | None,
    variables: NDArray[np.float64],
    functions: NDArray[np.float64],
    perturbed_variables: NDArray[np.float64],
    perturbed_functions: NDArray[np.float64],
    realization_weights: NDArray[np.float64] | None,
    failed_realizations: NDArray[np.bool_],
) -> NDArray[np.float64]:
    gradients = np.zeros((functions.shape[-1], variables.shape[-1]), dtype=np.float64)
    delta_variables = perturbed_variables - np.expand_dims(variables, axis=1)
    delta_functions = perturbed_functions - np.expand_dims(functions, axis=1)

    if estimator_indices is None:
        estimator_indices = np.zeros(functions.shape[1], dtype=np.intc)

    for estimator_idx, estimator in enumerate(function_estimators):
        mask = estimator_indices == estimator_idx
        for idx in np.where(mask)[0]:
            gradients[idx, ...] = _calculate_gradient(
                functions[..., idx],
                delta_variables,
                delta_functions[..., idx],
                failed_realizations,
                (
                    config.realizations.weights
                    if realization_weights is None
                    else realization_weights[idx, ...]
                ),
                estimator,
                merge_realizations=config.gradient.merge_realizations,
            )

    return gradients


def _calculate_estimated_objective_gradients(  # noqa: PLR0913
    config: EnOptConfig,
    function_estimators: list[FunctionEstimator],
    variables: NDArray[np.float64],
    functions: NDArray[np.float64],
    perturbed_variables: NDArray[np.float64],
    perturbed_functions: NDArray[np.float64],
    realization_weights: NDArray[np.float64] | None,
    failed_realizations: NDArray[np.bool_],
) -> NDArray[np.float64]:
    return _calculate_estimated_gradients(
        config,
        function_estimators,
        config.objectives.function_estimators,
        variables,
        functions,
        perturbed_variables,
        perturbed_functions,
        realization_weights,
        failed_realizations,
    )


def _calculate_estimated_constraint_gradients(  # noqa: PLR0913
    config: EnOptConfig,
    function_estimators: list[FunctionEstimator],
    variables: NDArray[np.float64],
    constraints: NDArray[np.float64] | None,
    perturbed_variables: NDArray[np.float64],
    perturbed_constaints: NDArray[np.float64] | None,
    realization_weights: NDArray[np.float64] | None,
    failed_realizations: NDArray[np.bool_],
) -> NDArray[np.float64] | None:
    if constraints is None:
        return None
    assert perturbed_constaints is not None
    assert config.nonlinear_constraints is not None
    return _calculate_estimated_gradients(
        config,
        function_estimators,
        config.nonlinear_constraints.function_estimators,
        variables,
        constraints,
        perturbed_variables,
        perturbed_constaints,
        realization_weights,
        failed_realizations,
    )



"""This module defines the abstract base class for optimization steps."""

from __future__ import annotations

import numpy as np

from ropt.results import FunctionResults, Results


def _get_new_optimal_result(
    optimal_result: FunctionResults | None, results: FunctionResults
) -> FunctionResults | None:
    if optimal_result is None:
        return results
    assert optimal_result.functions is not None
    assert results.functions is not None
    optimal = optimal_result.functions.weighted_objective
    objective = results.functions.weighted_objective
    if objective < optimal:
        return results
    return None


def _violates_constraint(results: Results, tolerance: float | None) -> bool:
    if tolerance is None:
        return False

    assert isinstance(results, FunctionResults)
    if results.constraint_info is None:
        return False

    for violations in (
        results.constraint_info.bound_violation,
        results.constraint_info.linear_violation,
        results.constraint_info.nonlinear_violation,
    ):
        if violations is not None and np.any(violations > tolerance):
            return True

    return False


def _get_last_result(
    results: tuple[Results, ...],
    constraint_tolerance: float | None,
) -> FunctionResults | None:
    return next(
        (
            item
            for item in reversed(results)
            if (
                isinstance(item, FunctionResults)
                and item.functions is not None
                and not _violates_constraint(item, constraint_tolerance)
            )
        ),
        None,
    )


def _update_optimal_result(
    optimal_result: FunctionResults | None,
    results: tuple[Results, ...],
    constraint_tolerance: float | None,
) -> FunctionResults | None:
    return_result: FunctionResults | None = None
    for item in results:
        if (
            isinstance(item, FunctionResults)
            and item.functions is not None
            and not _violates_constraint(item, constraint_tolerance)
        ):
            assert isinstance(item, FunctionResults)
            new_optimal_result = _get_new_optimal_result(optimal_result, item)
            if new_optimal_result is not None:
                optimal_result = new_optimal_result
                return_result = new_optimal_result
    return return_result



"""This module implements an optimization plugin that employs an external processes."""

from __future__ import annotations

import atexit
import contextlib
import json
import os
import selectors
import signal
import subprocess
import sys
import time
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Any, Final, Self

import numpy as np

from ropt.config.enopt import EnOptConfig
from ropt.exceptions import ConfigError, OptimizationAborted
from ropt.plugins import PluginManager

from .base import Optimizer, OptimizerCallback, OptimizerPlugin

if TYPE_CHECKING:
    from numpy.typing import NDArray


_PLUGIN_BINARY: Final = "ropt_plugin_optimizer"
_PROCESS_TIMEOUT: Final = 10


class ExternalOptimizer(Optimizer):
    """Plugin class for optimization using an external process.

    This class enables optimization via an external process, which performs the
    optimization independently and communicates with this class over pipes to
    request function evaluations, report optimizer states, and handle any
    errors.

    Typically, the optimizer is specified within an
    [`OptimizerConfig`][ropt.config.enopt.OptimizerConfig] via the `method`
    field, which either provides the algorithm name directly or follows the form
    `plugin-name/method-name`. In the first case, `ropt` searches among all
    available optimizer plugins to find the specified method. In the second
    case, it checks if the plugin identified by `plugin-name` contains
    `method-name` and, if so, uses it. Both of these are not supported by the
    external optimizer class. Instead, it requires that the `method` field
    includes both the plugin and method names in the format
    `external/plugin-name/method-name` or `external/method-name`. This ensures
    the external optimizer can identify and launch the specified optimization
    method `method-name` and launch it as an external process.
    """

    def __init__(
        self, config: EnOptConfig, optimizer_callback: OptimizerCallback
    ) -> None:
        """Initialize the optimizer.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        self._config = config
        self._optimizer_callback = optimizer_callback
        self._process_pid: int | None = None

        optimizer: Optimizer = (
            PluginManager()
            .get_plugin("optimizer", config.optimizer.method.split("/", maxsplit=1)[1])
            .create(config, lambda *_: None)
        )
        self._allow_nan = optimizer.allow_nan
        self._is_parallel = optimizer.is_parallel
        del optimizer

    def start(self, initial_values: NDArray[np.float64]) -> None:
        """Start the optimization.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        atexit.register(self._atexit)

        with TemporaryDirectory() as fifo_dir:
            fifo1 = Path(fifo_dir) / "fifo1"
            fifo2 = Path(fifo_dir) / "fifo2"

            with _JSONPipeCommunicator(fifo1, fifo2) as comm:
                try:
                    process = subprocess.Popen(  # noqa: S603
                        [_PLUGIN_BINARY, str(fifo2), str(fifo1), str(os.getpid())]
                    )
                    self._process_pid = process.pid
                except FileNotFoundError as exc:
                    msg = "The plugin runner binary was not found"
                    raise ConfigError(msg) from exc

                answer: str | list[Any] | dict[str, Any] | None = None
                exception: BaseException | None = None

                while process.poll() is None:
                    if answer is None:
                        try:
                            answer = self._handle_request(comm, initial_values)
                        except Exception as exc:  # noqa: BLE001
                            # Store the exception, we first need to send the 'abort' signal:
                            exception = exc
                            answer = "abort"

                    if answer is not None and comm.write(answer):
                        answer = None
                        # If the message has been sent, then reraise any exceptions:
                        if exception is not None:
                            # The process should have aborted:
                            with contextlib.suppress(ProcessLookupError):
                                os.kill(self._process_pid, signal.SIGTERM)
                            with contextlib.suppress(subprocess.TimeoutExpired):
                                process.wait(_PROCESS_TIMEOUT)
                            raise exception
                    time.sleep(0.1)

                with contextlib.suppress(ProcessLookupError):
                    os.kill(self._process_pid, signal.SIGTERM)
                with contextlib.suppress(subprocess.TimeoutExpired):
                    process.wait(_PROCESS_TIMEOUT)

    @property
    def allow_nan(self) -> bool:
        """Whether NaN is allowed.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        return self._allow_nan

    @property
    def is_parallel(self) -> bool:
        """Whether the current run is parallel.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        return self._is_parallel

    def _handle_request(
        self, comm: _JSONPipeCommunicator, initial_values: NDArray[np.float64]
    ) -> str | list[Any] | dict[str, Any] | None:
        request = comm.read()

        if request == "config":
            return self._config.model_dump(round_trip=True)

        if request == "initial_values":
            return initial_values.tolist()  # type: ignore[no-any-return]

        if isinstance(request, dict):
            if (evaluation := request.get("evaluation")) is not None:
                functions, gradients = self._optimizer_callback(
                    np.array(evaluation["variables"], dtype=np.float64),
                    return_functions=evaluation["return_functions"],
                    return_gradients=evaluation["return_gradients"],
                )
                return {
                    "functions": functions.tolist(),
                    "gradients": gradients.tolist(),
                }
            if (error := request.get("error")) is not None:
                msg = f"External optimizer error: {error}"
                raise RuntimeError(msg)

        return None

    def _atexit(self) -> None:
        if self._process_pid is not None:
            with contextlib.suppress(ProcessLookupError):
                os.kill(self._process_pid, signal.SIGTERM)


class ExternalOptimizerPlugin(OptimizerPlugin):
    """The external optimizer plugin class."""

    @classmethod
    def create(
        cls, config: EnOptConfig, optimizer_callback: OptimizerCallback
    ) -> ExternalOptimizer:
        """Initialize the optimizer plugin.

        See the [ropt.plugins.optimizer.base.OptimizerPlugin][] abstract base class.

        # noqa
        """
        return ExternalOptimizer(config, optimizer_callback)

    @classmethod
    def is_supported(cls, method: str) -> bool:
        """Check if a method is supported.

        See the [ropt.plugins.base.Plugin][] abstract base class.

        # noqa
        """
        return PluginManager().is_supported("optimizer", method)

    @classmethod
    def allows_discovery(cls) -> bool:
        """Check if the plugin can be discovered automatically.

        See the [ropt.plugins.optimizer.base.OptimizerPlugin][] abstract base class.

        # noqa
        """
        return False

    @classmethod
    def validate_options(
        cls, method: str, options: dict[str, Any] | list[str] | None
    ) -> None:
        """Validate the options of a given method.

        See the [ropt.plugins.optimizer.base.OptimizerPlugin][] abstract base class.

        # noqa
        """
        method = method.split("/", maxsplit=1)[1]
        PluginManager().get_plugin("optimizer", method).validate_options(
            method, options
        )


class _PluginOptimizer:
    def __init__(self, parent_pid: int) -> None:
        self._parent_pid = parent_pid
        self._comm: _JSONPipeCommunicator

    def _check_parent(self) -> None:
        try:
            os.kill(self._parent_pid, 0)
        except OSError:
            sys.exit(0)

    def _request(self, request: Any) -> Any:  # noqa: ANN401
        while True:
            self._check_parent()
            if self._comm.write(request):
                break
        while True:
            self._check_parent()
            answer = self._comm.read()
            if answer is not None:
                return answer

    def _callback(
        self,
        variables: NDArray[np.float64],
        *,
        return_functions: bool,
        return_gradients: bool,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        response = self._request(
            {
                "evaluation": {
                    "variables": variables.tolist(),
                    "return_functions": return_functions,
                    "return_gradients": return_gradients,
                }
            },
        )
        if response == "abort":
            raise OptimizationAborted(response)
        return (
            np.array(response["functions"], dtype=np.float64),
            np.array(response["gradients"], dtype=np.float64),
        )

    def run(self, fifo1: Path, fifo2: Path) -> int:
        with _JSONPipeCommunicator(fifo1, fifo2) as self._comm:
            config = EnOptConfig.model_validate(self._request("config"))

            optimizer = (
                PluginManager()
                .get_plugin(
                    "optimizer", config.optimizer.method.split("/", maxsplit=1)[1]
                )
                .create(config, self._callback)
            )
            try:
                optimizer.start(
                    np.array(self._request("initial_values"), dtype=np.float64)
                )
            except OptimizationAborted:
                return 0
            except Exception as exc:  # noqa: BLE001
                assert self._request({"error": str(exc)}) == "abort"  # noqa: PT017
                return 1
        return 0


def ropt_plugin_optimizer() -> int:
    """Entry point for the plugin runner script.

    Returns:
        The exit code of the script.
    """
    fifo1 = Path(sys.argv[1])
    fifo2 = Path(sys.argv[2])
    parent_pid = int(sys.argv[3])

    assert fifo1.exists()
    assert fifo2.exists()

    return _PluginOptimizer(parent_pid).run(fifo1, fifo2)


class _JSONPipeCommunicator:
    DELIMITER = "--READY--"

    def __init__(self, read_pipe: Path, write_pipe: Path, timeout: float = 1.0) -> None:
        self._read_pipe = read_pipe
        self._write_pipe = write_pipe
        self._timeout = timeout

        self._read_fd: int
        self._write_fd: int | None
        self._selector: selectors.BaseSelector

        if not read_pipe.exists():
            os.mkfifo(read_pipe)
        if not write_pipe.exists():
            os.mkfifo(write_pipe)

    def __enter__(self) -> Self:
        self._read_fd = os.open(self._read_pipe, os.O_RDONLY | os.O_NONBLOCK)
        self._write_fd = None
        self._selector = selectors.DefaultSelector()
        self._selector.register(self._read_fd, selectors.EVENT_READ)
        return self

    def __exit__(self, *args: object) -> None:
        self._selector.close()
        os.close(self._read_fd)
        if self._write_fd is not None:
            os.close(self._write_fd)

    def read(self) -> str | list[Any] | dict[str, Any] | None:
        events = self._selector.select(timeout=self._timeout)
        for _, mask in events:
            if mask & selectors.EVENT_READ:
                with os.fdopen(os.dup(self._read_fd), "r", encoding="utf-8") as fd:
                    buffer = ""
                    while line := fd.readline():
                        if line.strip() == self.DELIMITER:
                            buffer = buffer.strip()
                            return json.loads(buffer) if buffer else buffer
                        buffer += line
        return None

    def write(self, data: str | list[Any] | dict[str, Any]) -> bool:
        class CustomEncoder(json.JSONEncoder):
            def default(self, obj: Any) -> Any:  # noqa: ANN401
                if isinstance(obj, np.ndarray):
                    return obj.tolist()
                if isinstance(obj, Path):
                    return str(obj)
                return super().default(obj)

        if self._write_fd is None:
            self._write_fd = os.open(self._write_pipe, os.O_WRONLY | os.O_NONBLOCK)
            self._selector.register(self._write_fd, selectors.EVENT_WRITE)
        events = self._selector.select(timeout=self._timeout)
        for _, mask in events:
            if mask & selectors.EVENT_WRITE:
                os.write(
                    self._write_fd,
                    f"{json.dumps(data, cls=CustomEncoder)}\n{self.DELIMITER}\n".encode(),
                )
                return True
        return False
